#include <resources/alpath.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>


/*
 * mangled path format:
 * <type>://(<group>/)<name>:<path>
 *
 * e.g. res://test:/tests/test.png
 * type: "res"
 * group: ""
 * name: "test"
 * path: "/tests/test.png"
 *
 *
 * demangled path format:
 * ./<type_path>/<demangled_path>
 *
 * e.g.
 * type_path: res -> ./assets/
 * demangled_path: tests/test.png
 * -> ./assets/tests/test.png
 *
 * type max length: 4
 *
 * smallest possible path: ://:a (5 chars)
 *
 * groups max length: 26
 * groups:
 * e.g. res://image/test:/tests/test.png
 * type: "res"
 * group: "image"
 * name: "test"
 * path: "/tests/test.png"
 * -> ./assets/tests/test.png
 */


// Path

ALPath al_demangle_path(const char* path)
{

    const ALPath res = malloc(sizeof(alyson_path_t));
    res->type[0] = '\0';
    res->group[0] = '\0';
    res->name = NULL;
    res->path = NULL;

    const usx len = strlen(path);
    usx back_ptr = 0;

    if (len < 5)
    {
        fprintf(stderr, "[ALResource] Invalid path: \"%s\"; too short\n", path);
        al_destroy_path(res);
        return NULL;
    }

    // type
    for (usx i = 0; i < sizeof(res->type); i++)
    {
        if (path[i] == ':')
        {
            res->type[i] = '\0';
            back_ptr = i + 1;
            break;
        }
        res->type[i] = path[i];
    }
    if (back_ptr == 0)
    {
        fprintf(stderr, "[ALResource] Invalid path: \"%s\"; type not found\n", path);
        al_destroy_path(res);
        return NULL;
    }

    if (path[back_ptr] != '/' || path[back_ptr+1] != '/') {
        fprintf(stderr, "[ALResource] Invalid path: \"%s\"; type not followed by '://'\n", path);
        al_destroy_path(res);
        return NULL;
    }
    back_ptr += 2;

    // group + name
    for (usx i = back_ptr; i < len; i++)
    {
        if (path[i] == '/')
        {
            if (res->group[0] != '\0')
            {
                fprintf(stderr, "[ALResource] Invalid path: \"%s\"; group already set\n", path);
                al_destroy_path(res);
                return NULL;
            }
            if (back_ptr == i) break;
            strncpy(res->group, (char*)path + back_ptr, i - back_ptr);
            res->group[i-back_ptr] = '\0';
            back_ptr = i + 1;
        }

        if (path[i] == ':')
        {
            res->name = malloc(sizeof(char) * (i - back_ptr + 1));
            strncpy(res->name, (char*)path + back_ptr, i - back_ptr);
            res->name[i - back_ptr] = '\0';
            back_ptr = i + 1;
            break;
        }
    }

    // path
    if (back_ptr == len)
    {
        fprintf(stderr, "[ALResource] Invalid path: \"%s\"; path not found\n", path);
        al_destroy_path(res);
        return NULL;
    }
    res->path = malloc(sizeof(char) * (len - back_ptr + 1));
    strncpy(res->path, (char*)path + back_ptr, len - back_ptr);
    res->path[len - back_ptr] = '\0';

    return res;
}

char* al_mangle_path(const ALPath path)
{
    char* str = malloc(sizeof(char) * (
        strlen(path->type) +
        strlen(path->group) +
        strlen(path->name) +
        strlen(path->path) +
        5 ));
    if (path->group[0] != '\0')
    {
        sprintf(str, "%s://%s/%s:%s", path->type, path->group, path->name, path->path);
    }
    else
    {
        sprintf(str, "%s://%s:%s", path->type, path->name, path->path);
    }
    return str;
}

ALPath al_create_path(const char* type, const char* path, const char* name, const char* group)
{
    const ALPath res = malloc(sizeof(alyson_path_t));
    res->type[0] = '\0';
    res->name = NULL;
    res->path = NULL;

    if (type != NULL)
    {
        if (strlen(type) > 4)
        {
            fprintf(stderr, "[ALResource] Type name too long: %s\n", type);
            return NULL;
        }
        strncpy(res->type, type, strlen(type));
        res->type[4] = '\0';
    }

    if (path != NULL)
    {
        usx len = strlen(path);
        res->path = malloc(sizeof(char) * (len + 1));
        strncpy(res->path, path, len);
        res->path[len] = '\0';
    }

    if (name != NULL)
    {
        res->name = malloc(sizeof(char) * (strlen(name) + 1));
        strncpy(res->name, name, strlen(name));
        res->name[strlen(name)] = '\0';
    }

    if (group != NULL)
    {
        if (strlen(group) > 26)
        {
            fprintf(stderr, "[ALResource] Group name too long: %s\n", group);
            al_destroy_path(res);
            return NULL;
        }
        strncpy(res->group, group, strlen(group));
        res->group[strlen(group)] = '\0';
    }

    return res;
}

void al_destroy_path(ALPath path)
{
    if (path->name != NULL)
    {
        free(path->name);
        path->name = NULL;
    }
    if (path->path != NULL)
    {
        free(path->path);
        path->path = NULL;
    }
    free(path);
    path = NULL;
}
